#!/usr/bin/env python

import os
import re
import socket
import subprocess
import sys
import yaml
import base64
import grp
import pwd
import shutil
import os.path

from charmhelpers.core.hookenv import (
    open_port,
    close_port,
    log,
    config as config_get,
    relations_of_type,
)
from charmhelpers.contrib.charmsupport import nrpe

###############################################################################
# Global variables
###############################################################################
default_apache2_service_config_dir = "/var/run/apache2"
service_affecting_packages = ['apache2']
default_apache22_config_dir = "/etc/apache2/conf.d"
default_apache24_config_dir = "/etc/apache2/conf-available"

juju_warning_header = """#
#    "             "
#  mmm   m   m   mmm   m   m
#    #   #   #     #   #   #
#    #   #   #     #   #   #
#    #   "mm"#     #   "mm"#
#    #             #
#  ""            ""
# This file is managed by Juju. Do not make local changes.
#"""


###############################################################################
# Supporting functions
###############################################################################

#------------------------------------------------------------------------------
# apt_get_install( package ):  Installs a package
#------------------------------------------------------------------------------
def apt_get_install(packages=None):
    if packages is None:
        return False
    cmd_line = ['apt-get', '-y', 'install', '-qq']
    cmd_line.append(packages)
    return subprocess.call(cmd_line)


def ensure_package_status(packages, status):
    if status in ['install', 'hold']:
        selections = ''.join(['{} {}\n'.format(package, status)
                              for package in packages])
        dpkg = subprocess.Popen(['dpkg', '--set-selections'],
                                stdin=subprocess.PIPE)
        dpkg.communicate(input=selections)


#------------------------------------------------------------------------------
# apt_get_purge( package ):  Purges a package
#------------------------------------------------------------------------------
def apt_get_purge(packages=None):
    if packages is None:
        return False
    cmd_line = ['apt-get', '-y', 'purge', '-qq']
    cmd_line.append(packages)
    return subprocess.call(cmd_line)


#------------------------------------------------------------------------------
# service_apache2:  Convenience function to start/stop/restart/reload
#                   the apache2 service
#------------------------------------------------------------------------------
def service_apache2(action=None):
    if action is None:
        return
    elif action == "check":
        args = ['/usr/sbin/apache2ctl', 'configtest']
    else:
        args = ['service', 'apache2', action]
    ret_val = subprocess.call(args)
    return ret_val == 0


def run(command, *args, **kwargs):
    try:
        output = subprocess.check_output(command, *args, **kwargs)
        return output
    except Exception, e:
        print e
        raise


def enable_module(module=None):
    if module is None:
        return True
    if os.path.exists("/etc/apache2/mods-enabled/%s.load" % (module)):
        log("Module already loaded: %s" % module)
        return True
    if not os.path.exists("/etc/apache2/mods-available/%s.load" % (module)):
        return_value = apt_get_install("libapache2-mod-%s" % (module))
        if return_value != 0:
            log("Installing module %s failed" % (module))
            return False
    return_value = subprocess.call(['/usr/sbin/a2enmod', module])
    if return_value != 0:
        return False
    if service_apache2("check"):
        service_apache2("reload")
        return True


def disable_module(module=None):
    if module is None:
        return True
    if not os.path.exists("/etc/apache2/mods-enabled/%s.load" % (module)):
        log("Module already disabled: %s" % module)
        return True
    return_value = subprocess.call(['/usr/sbin/a2dismod', module])
    if return_value != 0:
        return False
    if service_apache2("check"):
        service_apache2("reload")
        return True


def is_apache24():
    return os.path.exists("/usr/sbin/a2enconf")


def site_filename(name, enabled=False):
    if enabled:
        sites_dir = "/etc/apache2/sites-enabled"
    else:
        sites_dir = "/etc/apache2/sites-available"

    if is_apache24():
        return "{}/{}.conf".format(sites_dir, name)
    else:
        return "{}/{}".format(sites_dir, name)


def conf_filename(name):
    """Return an apache2 config filename path, as:
      2.4: /etc/apache2/conf-available/foo.conf
      2.2: /etc/apache2/conf.d/foo
    """
    if is_apache24():
        return "{}/{}.conf".format(default_apache24_config_dir, name)
    else:
        return "{}/{}".format(default_apache22_config_dir, name)


def conf_enable(name):
    "Enable apache2 config without reloading service"
    if is_apache24():
        return subprocess.call(['/usr/sbin/a2enconf', name]) == 0
    # no-op otherwise
    return True


def conf_disable(name):
    "Disable apache2 config without reloading service"
    if is_apache24():
        return subprocess.call(['/usr/sbin/a2disconf', name]) == 0
    # no-op otherwise
    return True


###############################################################################
# Hook functions
###############################################################################
def install_hook():
    if not os.path.exists(default_apache2_service_config_dir):
        os.mkdir(default_apache2_service_config_dir, 0600)
    apt_get_install("python-jinja2")
    install_status = apt_get_install("apache2")
    if install_status == 0:
        ensure_package_status(service_affecting_packages,
                              config_get('package_status'))
    ensure_extra_packages()
    return install_status


def ensure_extra_packages():
    extra = str(config_get('extra_packages'))
    if extra:
        install_status = apt_get_install(extra)
        if install_status == 0:
            ensure_package_status(extra, config_get('package_status'))


def dump_data(data2dump, log_prefix):
    log_file = '/tmp/pprint-%s.log' % (log_prefix)
    if data2dump is not None:
        logFile = open(log_file, 'w')
        import pprint
        pprint.pprint(data2dump, logFile)
        logFile.close()


def get_reverseproxy_data(relation='reverseproxy'):
    relation_data = relations_of_type(relation)
    reverseproxy_data = {}
    if relation_data is None or len(relation_data) == 0:
        return reverseproxy_data
    for unit_data in relation_data:
        unit_name = unit_data["__unit__"]
        if 'port' not in unit_data:
            return reverseproxy_data
        # unit_name: <service-name>-<unit_number>
        # jinja2 templates require python-type variables, remove all characters
        # that do not comply
        unit_type = re.sub(r'(.*)/[0-9]*', r'\1', unit_name)
        unit_type = re.sub('[^a-zA-Z0-9_]*', '', unit_type)
        log('unit_type: %s' % unit_type)

        host = unit_data['private-address']
        if unit_type in reverseproxy_data:
            continue
        for config_setting in unit_data.keys():
            if config_setting in ("__unit__", "__relid__"):
                continue
            config_key = '%s_%s' % (unit_type,
                                    config_setting.replace("-", "_"))
            config_key = re.sub('[^a-zA-Z0-9_]*', '', config_key)
            reverseproxy_data[config_key] = unit_data[
                config_setting]
            reverseproxy_data[unit_type] = '%s:%s' % (
                host, unit_data['port'])
        if 'all_services' in unit_data:
            service_data = yaml.safe_load(unit_data['all_services'])
            for service_item in service_data:
                service_name = service_item['service_name']
                service_port = service_item['service_port']
                service_key = '%s_%s' % (unit_type, service_name)
                service_key = re.sub('[^a-zA-Z0-9_]*', '', service_key)
                reverseproxy_data[service_key] = '%s:%s' % (host, service_port)
    return reverseproxy_data


def update_balancers():
    relation_data = relations_of_type('balancer')
    if relation_data is None or len(relation_data) == 0:
        log("No relation data, exiting.")
        return

    unit_dict = {}
    for unit_data in relation_data:
        unit_name = unit_data["__unit__"]
        if "port" not in unit_data:
            log("No port in relation data for '%s', skipping." % unit_name)
            continue
        port = unit_data["port"]
        if "private-address" not in unit_data:
            log("No private-address in relation data for '%s', skipping." %
                unit_name)
            continue
        host = unit_data['private-address']

        if "all_services" in unit_data:
            service_data = yaml.safe_load(unit_data[
                "all_services"])
            for service_item in service_data:
                service_port = service_item["service_port"]
                current_units = unit_dict.setdefault(
                    service_item["service_name"], [])
                current_units.append("%s:%s" % (host, service_port))
        else:
            if "sitenames" in unit_data:
                unit_types = unit_data["sitenames"].split()
            else:
                unit_types = (re.sub(r"(.*)/[0-9]*", r"\1", unit_name),)

            for unit_type in unit_types:
                current_units = unit_dict.setdefault(unit_type, [])
                current_units.append("%s:%s" % (host, port))

    if not unit_dict:
        return

    write_balancer_config(unit_dict)
    return unit_dict


def write_balancer_config(unit_dict):
    config_data = config_get()

    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(
        os.environ['CHARM_DIR'], 'data')))
    for balancer_name in unit_dict.keys():
        balancer_host_file = conf_filename('{}.balancer'.format(balancer_name))
        templ_vars = {
            'balancer_name': balancer_name,
            'balancer_addresses': unit_dict[balancer_name],
            'lb_balancer_timeout': config_data['lb_balancer_timeout'],
        }
        template = template_env.get_template(
            'balancer.template').render(templ_vars)
        log("Writing file: %s with data: %s" % (balancer_host_file,
                                                templ_vars))
        with open(balancer_host_file, 'w') as balancer_config:
            balancer_config.write(str(template))
        conf_enable('{}.balancer'.format(balancer_name))


def update_nrpe_checks():
    nrpe_compat = nrpe.NRPE()
    conf = nrpe_compat.config
    check_http_params = conf.get('nagios_check_http_params')
    if check_http_params:
        nrpe_compat.add_check(
            shortname='vhost',
            description='Check Virtual Host',
            check_cmd='check_http %s' % check_http_params
        )
    nrpe_compat.write()


def create_mpm_workerfile():
    config_data = config_get()
    mpm_workerfile = conf_filename('000mpm-worker')
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(
        os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'mpm_type': config_data['mpm_type'],
        'mpm_startservers': config_data['mpm_startservers'],
        'mpm_minsparethreads': config_data['mpm_minsparethreads'],
        'mpm_maxsparethreads': config_data['mpm_maxsparethreads'],
        'mpm_threadlimit': config_data['mpm_threadlimit'],
        'mpm_threadsperchild': config_data['mpm_threadsperchild'],
        'mpm_serverlimit': config_data['mpm_serverlimit'],
        'mpm_maxclients': config_data['mpm_maxclients'],
        'mpm_maxrequestsperchild': config_data['mpm_maxrequestsperchild'],
    }
    template = \
        template_env.get_template('mpm_worker.template').render(templ_vars)
    with open(mpm_workerfile, 'w') as mpm_config:
        mpm_config.write(str(template))
    conf_enable('000mpm-worker')


def create_security():
    config_data = config_get()
    securityfile = conf_filename('security')
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(
        os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'juju_warning_header': juju_warning_header,
        'server_tokens': config_data['server_tokens'],
        'server_signature': config_data['server_signature'],
        'trace_enabled': config_data['trace_enabled'],
        'ssl_protocol': config_data['ssl_protocol'],
        'ssl_honor_cipher_order': config_data['ssl_honor_cipher_order'],
        'ssl_cipher_suite': config_data['ssl_cipher_suite'],
        'is_apache24': is_apache24(),
    }
    template = \
        template_env.get_template('security.template').render(templ_vars)
    with open(securityfile, 'w') as security_config:
        security_config.write(str(template))
    conf_enable('security')


def ship_logrotate_conf():
    config_data = config_get()
    logrotate_file = '/etc/logrotate.d/apache2'
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(
        os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'juju_warning_header': juju_warning_header,
        'logrotate_rotate': config_data['logrotate_rotate'],
        'logrotate_count': config_data['logrotate_count'],
        'logrotate_dateext': config_data['logrotate_dateext'],
    }
    template = template_env.get_template('logrotate.conf.template').render(
        templ_vars)
    with open(logrotate_file, 'w') as logrotate_conf:
        logrotate_conf.write(str(template))


def config_changed():
    relationship_data = {}
    config_data = config_get()

    ensure_package_status(service_affecting_packages,
                          config_data['package_status'])
    ensure_extra_packages()

    relationship_data.update(get_reverseproxy_data(relation='reverseproxy'))
    relationship_data.update(get_reverseproxy_data(relation='website-cache'))
    if update_balancers():
        # apache 2.4 has lbmethods split, needs to enable specific module(s)
        if is_apache24():
            enable_module('lbmethod_byrequests')

    if config_data['enable_modules']:
        module_list = config_data['enable_modules'].split()
        for module in module_list:
            enable_module(module)

    if config_data['disable_modules']:
        module_list = config_data['disable_modules'].split()
        for module in module_list:
            disable_module(module)

    if config_data['mpm_type'] == 'worker':
        apt_get_install('apache2-mpm-worker')
        apt_get_purge('apache2-mpm-prefork')
    elif config_data['mpm_type'] == 'prefork':
        apt_get_install('apache2-mpm-prefork')
        apt_get_purge('apache2-mpm-worker')
    create_mpm_workerfile()
    create_security()
    ports = {'http': 80, 'https': 443}
    for proto in ports.keys():
        template_var = 'vhost_%s_template' % (proto)
        template_data = dict(config_data.items() + relationship_data.items())
        close_port(ports[proto])
        if template_var in config_data:
            vhost_name = '%s_%s' % (config_data['servername'], proto)
            vhost_file = site_filename(vhost_name)
            from jinja2 import Template
            template = Template(
                str(base64.b64decode(config_data[template_var])))
            log("Writing file: %s with data: %s" % (vhost_file, template_data))
            with open(vhost_file, 'w') as vhost:
                vhost.write(str(template.render(template_data)))
            open_port(ports[proto])
        subprocess.call(['/usr/sbin/a2ensite', vhost_name])

    cert_file = None
    if config_data['ssl_certlocation']:
        cert_file = '/etc/ssl/certs/%s' % \
            (config_data['ssl_certlocation'].rpartition('/')[2])
    key_file = None
    if config_data['ssl_keylocation']:
        key_file = '/etc/ssl/private/%s' % \
            (config_data['ssl_keylocation'].rpartition('/')[2])
    chain_file = None
    if config_data['ssl_chainlocation']:
        chain_file = '/etc/ssl/certs/%s' % \
            (config_data['ssl_chainlocation'].rpartition('/')[2])

    if cert_file is not None and key_file is not None:
        # ssl_cert is SELFSIGNED so generate self-signed certificate for use.
        if config_data['ssl_cert'] and config_data['ssl_cert'] == "SELFSIGNED":
            os.environ['OPENSSL_CN'] = config_data['servername']
            os.environ['OPENSSL_PUBLIC'] = run(["unit-get",
                                                "public-address"]).rstrip()
            os.environ['OPENSSL_PRIVATE'] = run(["unit-get",
                                                 "private-address"]).rstrip()
            run(['openssl', 'req', '-new', '-x509', '-nodes', '-config',
                 os.path.join(os.environ['CHARM_DIR'], 'data', 'openssl.cnf'),
                 '-keyout', key_file, '-out', cert_file])

        # Use SSL certificate and key provided either as a base64 string or
        # shipped out with the charm.
        else:
            # Certificate provided as base64-encoded string.
            if config_data['ssl_cert']:
                log("Writing cert from config ssl_cert: %s" % cert_file)
                with open(cert_file, 'w') as f:
                    f.write(str(base64.b64decode(config_data['ssl_cert'])))
            # Use certificate file shipped out with charm.
            else:
                source = os.path.join(os.environ['CHARM_DIR'], 'data',
                                      config_data['ssl_certlocation'])
                if os.path.exists(source):
                    shutil.copy(source, cert_file)
                else:
                    log("Certificate not found, ignoring: %s" % source)

            # Private key provided as base64-encoded string.
            if config_data['ssl_key']:
                log("Writing key from config ssl_key: %s" % key_file)
                with open(key_file, 'w') as f:
                    f.write(str(base64.b64decode(config_data['ssl_key'])))
            # Use private key shipped out with charm.
            else:
                source = os.path.join(os.environ['CHARM_DIR'], 'data',
                                      config_data['ssl_keylocation'])
                if os.path.exists(source):
                    shutil.copy(source, key_file)
                else:
                    log("Key file not found, ignoring: %s" % source)

            if chain_file is not None:
                # Chain certificates provided as base64-encoded string.
                if config_data['ssl_chain']:
                    log("Writing chain certificates file from"
                        "config ssl_chain: %s" % chain_file)
                    with open(chain_file, 'w') as f:
                        f.write(str(base64.b64decode(
                            config_data['ssl_chain'])))
                # Use chain certificates shipped out with charm.
                else:
                    source = os.path.join(os.environ['CHARM_DIR'], 'data',
                                          config_data['ssl_chainlocation'])
                    if os.path.exists(source):
                        shutil.copy(source, chain_file)
                    else:
                        log("Chain certificates not found, "
                            "ignoring: %s" % source)

        # Tighten permissions on private key file.
        if os.path.exists(key_file):
            os.chmod(key_file, 0440)
            os.chown(key_file, pwd.getpwnam('root').pw_uid,
                     grp.getgrnam('ssl-cert').gr_gid)

    apache_syslog_conf = conf_filename("syslog")
    rsyslog_apache_conf = "/etc/rsyslog.d/45-apache2.conf"
    if config_data['use_rsyslog']:
        shutil.copy2("data/syslog-apache.conf", apache_syslog_conf)
        conf_enable("syslog")
        shutil.copy2("data/syslog-rsyslog.conf", rsyslog_apache_conf)
        # Fix permissions of access.log and error.log to allow syslog user to
        # write to
        os.chown("/var/log/apache2/access.log", pwd.getpwnam('syslog').pw_uid,
                 pwd.getpwnam('syslog').pw_gid)
        os.chown("/var/log/apache2/error.log", pwd.getpwnam('syslog').pw_uid,
                 pwd.getpwnam('syslog').pw_gid)
    else:
        conf_disable("syslog")
        if os.path.exists(apache_syslog_conf):
            os.unlink(apache_syslog_conf)
        if os.path.exists(rsyslog_apache_conf):
            os.unlink(rsyslog_apache_conf)
    run(["/usr/sbin/service", "rsyslog", "restart"])

    # Disable the default website because we don't want people to see the
    # "It works!" page on production services and remove the
    # conf.d/other-vhosts-access-log conf.
    if os.path.exists(site_filename("000-default", True)):
        run(["/usr/sbin/a2dissite", "000-default"])
    conf_disable("other-vhosts-access-log")
    if os.path.exists(conf_filename("other-vhosts-access-log")):
        os.unlink(conf_filename("other-vhosts-access-log"))

    if service_apache2("check"):
        if config_data["config_change_command"] in ["reload", "restart"]:
            service_apache2(config_data["config_change_command"])

    if config_data['openid_provider']:
        if not os.path.exists('/etc/apache2/security'):
            os.mkdir('/etc/apache2/security', 0755)
        with open('/etc/apache2/security/allowed-ops.txt', 'w') as f:
            f.write(config_data['openid_provider'].replace(',','\n'))
            f.write('\n')
            os.chmod('/etc/apache2/security/allowed-ops.txt', 0444)

    update_nrpe_checks()
    ship_logrotate_conf()


def start_hook():
    if service_apache2("status"):
        return(service_apache2("restart"))
    else:
        return(service_apache2("start"))


def stop_hook():
    if service_apache2("status"):
        return(service_apache2("stop"))


def reverseproxy_interface(hook_name=None):
    if hook_name is None:
        return(None)
    if hook_name == "changed":
        config_changed()


def website_interface(hook_name=None):
    if hook_name is None:
        return(None)
    my_host = socket.getfqdn(socket.gethostname())
    if my_host == "localhost":
        my_host = socket.gethostname()
    default_port = 80
    subprocess.call(['relation-set', 'port=%d' % default_port,
                     'hostname=%s' % my_host])


###############################################################################
# Main section
###############################################################################
def main(hook_name):
    if hook_name == "install":
        install_hook()
    elif hook_name == "config-changed" or hook_name == "upgrade-charm":
        config_changed()
    elif hook_name == "start":
        start_hook()
    elif hook_name == "stop":
        stop_hook()
    elif hook_name == "reverseproxy-relation-broken":
        config_changed()
    elif hook_name == "reverseproxy-relation-changed":
        config_changed()
    elif hook_name == "reverseproxy-relation-joined":
        config_changed()
    elif hook_name == "balancer-relation-broken":
        config_changed()
    elif hook_name == "balancer-relation-changed":
        config_changed()
    elif hook_name == "balancer-relation-joined":
        config_changed()
    elif hook_name == "website-cache-relation-broken":
        config_changed()
    elif hook_name == "website-cache-relation-changed":
        config_changed()
    elif hook_name == "website-cache-relation-joined":
        config_changed()
    elif hook_name == "website-relation-joined":
        website_interface("joined")
    elif hook_name in ("nrpe-external-master-relation-changed",
                       "local-monitors-relation-changed"):
        update_nrpe_checks()
    else:
        print "Unknown hook"
        sys.exit(1)

if __name__ == "__main__":
    hook_name = os.path.basename(sys.argv[0])
    # Also support being invoked directly with hook as argument name.
    if hook_name == "hooks.py":
        if len(sys.argv) < 2:
            sys.exit("Missing required hook name argument.")
        hook_name = sys.argv[1]
    main(hook_name)
